#include <sourcemod>
#include <sdktools>

public Plugin:myinfo = 
{
	name = "L4D Homerank",
	author = "ArmGadge",
	description = "Display records when player got Tank's claw attack.",
	version = "1.1.2",
	url = "http://radiowave.moe.hm/"
};

/*
todo
*着地判定の正確さを上げるロジックの実装（加速度でチェック？）

りれき

1.1.2
フィナーレの脱出の瞬間に飛んでいる場合、プレイヤーのワープがかかって誤判定が起きるため、測定をやめる処理を追加。

1.1.1
ランクデータファイルが存在しない場合は新しく作るようにする処理を、プラグイン起動時だけでなく、ランクデータリロード時にも行うようにした。
ランクデータをファイルに書き込むタイミングを、即時ではなくラウンド終了時に変更。書き込み時にラグる時があるため。

1.1.0
ハング（＝プレイヤーが淵にぶら下がっている）状態でTankに殴られると、着地判定が出ないため、異常に長い滞空時間が記録されていた。
→記録しないように変更。他にも宙吊り状態になるような他のシチュエーションで殴られると起こるかも。smokerに捕まってるときとか。未対策。
l4d_homerank_rank_reloadコマンドのランキングデータの再読み込み動作が正しくなかったのを修正
バージョン番号を示すcvarを追加
メッセージの多言語対応化
ランク保持数が変更できるようになった
ランクデータのエスケープ文字が読めてなかったのを修正

既知の問題
ランクインのメッセージ表示で、文字の間になんか謎の空白が入る。多言語メッセージの扱い方が悪い？

1.0.0 初版
*/

new bool:gb_isplayerflying[MAXPLAYERS+1];
new Float:gf_hitpos[MAXPLAYERS+1][3];
new Float:gf_hitheight[MAXPLAYERS+1];
new Float:gf_hitgametime[MAXPLAYERS+1];
new Float:gf_maxheight[MAXPLAYERS+1];
//new Float:gf_pframeheight[MAXPLAYERS+1];
//new Float:gf_ongroundgtime[MAXPLAYERS+1];
new String:gs_mapname[64];
new String:gs_rankfilepath[PLATFORM_MAX_PATH];
new Handle:gh_cv_disp_cond = INVALID_HANDLE;
new Handle:gh_cv_disp_dist = INVALID_HANDLE;
new Handle:gh_cv_disp_height = INVALID_HANDLE;
new Handle:gh_cv_disp_time = INVALID_HANDLE;
new Handle:gh_cv_ranksize = INVALID_HANDLE;
new Handle:gh_rankkv = INVALID_HANDLE;
new bool:gb_needupdatefile = false;

public OnPluginStart()
{
	// Require Left 4 Dead (2)
	decl String:game_name[64];
	GetGameFolderName(game_name, sizeof(game_name));
	if (!StrEqual(game_name, "left4dead", false) 
		&& !StrEqual(game_name, "left4dead2", false))
	{
		SetFailState("Plugin supports Left 4 Dead or Left 4 Dead 2 only.");
	}	
	
	HookEvent("player_hurt", Ev_PlayerHurt);
	HookEvent("player_death", Ev_PlayerLand);
	HookEvent("player_ledge_grab", Ev_PlayerLand);
	HookEvent("round_end", Ev_RoundEnd);
	HookEvent("finale_vehicle_leaving", Ev_FinaleVehicleLeaving);
	
	decl String:s_pluginver[32];
	GetPluginInfo(INVALID_HANDLE, PlInfo_Version, s_pluginver, sizeof(s_pluginver));
	CreateConVar("l4d_homerank_version", s_pluginver, "Version of Homerank Plugin", FCVAR_REPLICATED|FCVAR_NOTIFY|FCVAR_DONTRECORD);	
	gh_cv_disp_cond = CreateConVar("l4d_homerank_disp_cond", "2", "Conditional expression of displaying player's record [0:off/1:and/2:or]",
						0, true, 0.0, true, 2.0);
	gh_cv_disp_dist = CreateConVar("l4d_homerank_disp_dist", "0.0", "Plugin will display player's record when the distance larger than this value. (float)[meters]",
						0, true, 0.0, false);
	gh_cv_disp_height = CreateConVar("l4d_homerank_disp_height", "0.0", "Plugin will display player's record when the height larger than this value. (float)[meters]",
						0, true, 0.0, false);
	gh_cv_disp_time = CreateConVar("l4d_homerank_disp_time", "0.0", "Plugin will display player's record when the time longer than this value. (float)[seconds]",
						0, true, 0.0, false);
	gh_cv_ranksize = CreateConVar("l4d_homerank_ranksize", "10", "Size(=max number) of Homerun-ranking.(int)", 0, true, 0.0, false);
	
	RegAdminCmd ("l4d_homerank_rank_reload", Cmd_RankReload, ADMFLAG_ROOT, "Reload ranking data.");
	RegAdminCmd ("l4d_homerank_rank_flush", Cmd_RankFlush, ADMFLAG_ROOT, "Force flush ranking data.");

	AutoExecConfig(true, "l4d_homerank");			
	BuildPath(Path_SM, gs_rankfilepath, sizeof(gs_rankfilepath), "data/l4d_homerank.txt");
	
	LoadTranslations("plugin.l4d_homerank");

	loadRankData();	
}

public OnPluginEnd()
{
	CloseHandle(gh_rankkv);
}

public OnMapStart()
{
	GetCurrentMap(gs_mapname, sizeof(gs_mapname));
	
	for (new i=1; i<=MaxClients; i++)
	{
		gb_isplayerflying[i] = false;
		gf_maxheight[i] = 0.0;
	}
}

public Ev_PlayerHurt(Handle:event, const String:name[], bool:dontBroadcast)
{
	new client = GetClientOfUserId(GetEventInt(event, "userid"));
	decl String:str_weapon[64];
	
	//ハング中は殴られても着地しないため除外
	if(IsClientInGame(client) && GetClientTeam(client) == 2 && !IsFakeClient(client) &&
		GetEntProp(client, Prop_Send, "m_isHangingFromLedge") == 0)
	{
		GetEventString(event, "weapon", str_weapon, sizeof(str_weapon));
		
		if (StrEqual(str_weapon, "tank_claw"))
		{
			gb_isplayerflying[client] = true;
			GetClientAbsOrigin(client, gf_hitpos[client]);
			
			//高さ用
			gf_hitheight[client] = gf_hitpos[client][2];
			gf_maxheight[client] = gf_hitheight[client];
			//平面距離で出すためz軸を0に固定
			gf_hitpos[client][2] = 0.0;
			
			gf_hitgametime[client] = GetGameTime();
		}
	}
}

public Ev_PlayerLand(Handle:event, const String:name[], bool:dontBroadcast)
{
	new client = GetClientOfUserId(GetEventInt(event, "userid"));
	if (gb_isplayerflying[client] == true)
	{
		gb_isplayerflying[client] = false;
		checkResult(client);
	}
}

public Ev_FinaleVehicleLeaving(Handle:event, const String:name[], bool:dontBroadcast)
{
	//フィナーレの脱出の瞬間に飛んでいる場合、プレイヤーのワープがかかって誤判定になるから、測定をやめる。
	for (new i=1; i<=MaxClients; i++)
		if (gb_isplayerflying[i] == true) gb_isplayerflying[i] = false;
}

public Ev_RoundEnd(Handle:event, const String:name[], bool:dontBroadcast)
{
	//セーブのタイミングをラウンドエンドに変更
	
	if (gb_needupdatefile)
	{
		saveRankData();
		gb_needupdatefile = false;
		PrintToServer("[Homerank] Updated rankfile!");
	}
	else PrintToServer("[Homerank] There is no update for rankfile.");
}

public OnGameFrame()
{
	new Float:curpos[3];
	//z軸移動量判定用（未使用）
	//new Float:f_curgtime = GetGameTime();
	//new Float:f_time_stayground;
	
	for (new i=1; i<=MaxClients; i++)
	{
		if (gb_isplayerflying[i] == true)
		{
			GetClientAbsOrigin(i, curpos);
			if (curpos[2] > gf_maxheight[i]) gf_maxheight[i] = curpos[2];
			//z軸の移動量を確認（未使用）
			/*
			new Float:delta_z = FloatAbs(curpos[2] - gf_pframeheight[i]);
			
			if (delta_z <= 0.5 / 0.01905)
			{
				f_time_stayground = f_curgtime - gf_ongroundgtime[i];
			}
			else
			{
				gf_ongroundgtime[i] = f_curgtime;
			}
			*/
			//debug
			//PrintToChat(i, "gtime:%f ongroundtime:%f df_z:%f", f_curgtime, f_time_stayground, delta_z);
			
			//if (GetEntPropEnt(i, Prop_Send, "m_hGroundEntity") != -1 || f_time_stayground >= 0.5)
			if (GetEntPropEnt(i, Prop_Send, "m_hGroundEntity") != -1)
			{
				gb_isplayerflying[i] = false;
				checkResult(i);
			}
			//z軸移動量判定用（未使用）
			//gf_pframeheight[i] = curpos[2];
		}
	}
}

checkResult(client)
{
	new rank, Float:f_dist_m, Float:f_height_m, Float:f_time;	
	//new bool:upd_result_dist, bool:upd_result_height, bool:upd_result_time;
	
	f_time = calcTime(client);
	
	//↓稀に瞬間的に着地した際にマイナス値が出るため対策
	if (f_time > 0.0)
	{
		f_dist_m = calcDistance(client);
		f_height_m = calcHeight(client);
		//ランクを検索
		rank = searchRank("distance", f_dist_m);
		//ランクインしているなら更新
		if (rank <= GetConVarInt(gh_cv_ranksize))
		{
			updateRank("distance", rank, client, f_dist_m);
		}
		
		//同様に、高さと時間も確認と更新をする
		rank = searchRank("height", f_height_m);
		if (rank <= GetConVarInt(gh_cv_ranksize))
		{
			updateRank("height", rank, client, f_height_m);
		}
		
		rank = searchRank("time", f_time);
		if (rank <= GetConVarInt(gh_cv_ranksize))
		{
			updateRank("time", rank, client, f_time);
		}
		
		/*
		upd_result_dist = updateRank(client, "distance", dist_m);
		upd_result_height = updateRank(client, "height", height_m);
		upd_result_time = updateRank(client, "time", time);
		*/
		
		if (GetConVarInt(gh_cv_disp_cond) == 1)
		{
			if ((f_dist_m >= GetConVarFloat(gh_cv_disp_dist)) &&
				(f_height_m >= GetConVarFloat(gh_cv_disp_height)) &&
				(f_time >= GetConVarFloat(gh_cv_disp_time)))
				displayPlayerRecord(client, f_dist_m, f_height_m, f_time);
		}
		else if (GetConVarInt(gh_cv_disp_cond) == 2)
		{
			if ((f_dist_m >= GetConVarFloat(gh_cv_disp_dist)) ||
				(f_height_m >= GetConVarFloat(gh_cv_disp_height)) ||
				(f_time >= GetConVarFloat(gh_cv_disp_time)))
				displayPlayerRecord(client, f_dist_m, f_height_m, f_time);
		}
	}

}

displayPlayerRecord(client, Float:distance, Float:height, Float:time)
{
	/*
	PrintToChat(client, "[Homerank]Your flight record; Distance: %.1fm / Height: %.1fm / Time: %.3fsec.",
		distance, height, time);
	*/
	PrintToChat(client, "[Homerank] %t",
		"yourrecord", distance, height, time);
}

Float:calcDistance(client)
{
	new Float:landpos[3], Float:dist;

	GetClientAbsOrigin(client, landpos);
	//平面距離で出すためz軸を0に固定
	landpos[2] = 0.0;
	dist = GetVectorDistance(gf_hitpos[client], landpos);
	dist = dist*0.01905;
	//小数第一位で出すために10倍して丸めて10で割る
	//dist_m　= RoundFloat(dist_m*10.0)/10.0;
	return dist;
}

Float:calcHeight(client)
{
	new Float:height;
	height = (gf_maxheight[client] - gf_hitheight[client]) * 0.01905;
	return height;
}

Float:calcTime(client)
{
	new Float:time;
	time = GetGameTime() - gf_hitgametime[client];
	return time;
}

searchRank(const String:s_ranktype[], Float:value)
{
	decl String:str_rankno[8];
	new rank;
	
	//ranksize＝ランキングの最大値
	new ranksize = GetConVarInt(gh_cv_ranksize);
	if (StrEqual(s_ranktype, "distance", false)) KvJumpToKey(gh_rankkv, "Distance", true);
	else if (StrEqual(s_ranktype, "height", false)) KvJumpToKey(gh_rankkv, "Height", true);
	else if (StrEqual(s_ranktype, "time", false)) KvJumpToKey(gh_rankkv, "Time", true);
	
	//何位に入るか検索　同じ値の場合は新しいものは下位に来る仕様
	
	for (new currank=ranksize; currank>=1; currank--)
	{
		IntToString(currank, str_rankno, sizeof(str_rankno));
		//PrintToServer("%s", str_rankno);
		if (KvJumpToKey(gh_rankkv, str_rankno))
		{
			if (value <= KvGetFloat(gh_rankkv, "Value"))
			{
				rank = currank+1;
				KvGoBack(gh_rankkv);
				break;
			}
			KvGoBack(gh_rankkv);
		}
	}
	//データが何も無いときはrankは0になるので1位にしとく
	if (rank == 0) rank = 1;
	KvRewind(gh_rankkv);
	
	return rank;
}

bool:updateRank(const String:ranktype[], rank, client, Float:value)
{
	decl String:str_rankno[8];

	decl String:buf_ranktype[32], String:buf_playername[64], String:buf_value[16], String:buf_mapname[64];
	new Float:buf_distance;

	//ranksize＝ランキングの最大値
	new ranksize = GetConVarInt(gh_cv_ranksize);
	
	if (StrEqual(ranktype, "distance", false)) KvJumpToKey(gh_rankkv, "Distance", true);
	else if (StrEqual(ranktype, "height", false)) KvJumpToKey(gh_rankkv, "Height", true);
	else if (StrEqual(ranktype, "time", false)) KvJumpToKey(gh_rankkv, "Time", true);

	//その順位までのランキングを一つ下にずらす
	//最下位であれば、単に最下位を書き換えるだけでOKなのでこの処理は必要ない
	if (rank <= ranksize - 1)
	{
		for (new currank=ranksize-1; currank>=rank; currank--)
		{
			IntToString(currank, str_rankno, sizeof(str_rankno));
			if (KvJumpToKey(gh_rankkv, str_rankno))
			{
				KvGetString(gh_rankkv, "PlayerName", buf_playername, sizeof(buf_playername));
				buf_distance = KvGetFloat(gh_rankkv, "Value");
				KvGetString(gh_rankkv, "MapName", buf_mapname, sizeof(buf_mapname));
				KvGoBack(gh_rankkv);
				
				IntToString(currank+1, str_rankno, sizeof(str_rankno));
				if (KvJumpToKey(gh_rankkv, str_rankno, true))
				{
					KvSetString(gh_rankkv, "PlayerName", buf_playername);
					KvSetFloat(gh_rankkv, "Value", buf_distance);
					KvSetString(gh_rankkv, "MapName", buf_mapname);
				}
				KvGoBack(gh_rankkv);
			}
		}
	}
	
	//ランクに書き込む
	IntToString(rank, str_rankno, sizeof(str_rankno));
	if (KvJumpToKey(gh_rankkv, str_rankno, true))
	{
		GetClientName(client, buf_playername, sizeof(buf_playername));
		KvSetString(gh_rankkv, "PlayerName", buf_playername);
		KvSetFloat(gh_rankkv, "Value", value);
		KvSetString(gh_rankkv, "MapName", gs_mapname);
	}		
	
	KvRewind(gh_rankkv);
	
	gb_needupdatefile = true;
	
	if (StrEqual(ranktype, "distance", false))
	{
		strcopy(buf_ranktype, sizeof(buf_ranktype), "type_distance");
		//Format(buf_ranktype, sizeof(buf_ranktype), "%T", "cat_distance", client);
		Format(buf_value, sizeof(buf_value), "%.1fm", value);
	}
	else if (StrEqual(ranktype, "height", false))
	{
		strcopy(buf_ranktype, sizeof(buf_ranktype), "type_height");
		//Format(buf_ranktype, sizeof(buf_ranktype), "%T", "cat_height", client);
		Format(buf_value, sizeof(buf_value), "%.1fm", value);
	}
	else if	(StrEqual(ranktype, "time", false))
	{
		strcopy(buf_ranktype, sizeof(buf_ranktype), "type_time");
		//Format(buf_ranktype, sizeof(buf_ranktype), "%T", "cat_time", client);
		Format(buf_value, sizeof(buf_value), "%.3fsec.", value);
	}
	//buf_ranktypeの中身によって翻訳メッセージを変えてる
	if (rank == 1)
		PrintToChatAll("\x01[Homerank][%t] \x04%t", buf_ranktype, "newrecord", buf_playername, buf_value);
	else PrintToChatAll("\x01[Homerank][%t] \x03%t", buf_ranktype, "rankin", buf_playername, rank, buf_value);
	//翻訳について %t複数使えるらしい
}
/*
bool:updateRank(client, const String:ranktype[], Float:value)
{
	decl String:str_rankno[8];
	new String:playername[64];
	
	decl String:buf_ranktype[32], String:buf_playername[64], String:buf_value[16], String:buf_mapname[64];
	new Float:buf_distance;
	new rank;
	
	//ranksize＝ランキングの最大値
	new ranksize = GetConVarInt(gh_cv_ranksize);
	GetClientName(client, playername, sizeof(playername));
	if (StrEqual(ranktype, "distance", false)) KvJumpToKey(gh_rankkv, "Distance", true);
	else if (StrEqual(ranktype, "height", false)) KvJumpToKey(gh_rankkv, "Height", true);
	else if (StrEqual(ranktype, "time", false)) KvJumpToKey(gh_rankkv, "Time", true);
	
	//何位に入るか検索　同じ値の場合は新しいものは下位に来る仕様
	
	for (new currank=ranksize; currank>=1; currank--)
	{
		IntToString(currank, str_rankno, sizeof(str_rankno));
		//PrintToServer("%s", str_rankno);
		if (KvJumpToKey(gh_rankkv, str_rankno))
		{
			if (value <= KvGetFloat(gh_rankkv, "Value"))
			{
				rank = currank+1;
				KvGoBack(gh_rankkv);
				break;
			}
			KvGoBack(gh_rankkv);
		}
	}
	//データが何も無いときはrankは0になるので1位にしとく
	if (rank == 0) rank = 1;
	
	//ランクサイズ+1位は番外だから処理しない
	if (rank == ranksize + 1)
	{
		KvRewind(gh_rankkv);
		return false;
	}
	else
	{

		
		

		return true;
	}
}
*/

loadRankData()
{
	gh_rankkv = CreateKeyValues("HomerunRanking");
	KvSetEscapeSequences(gh_rankkv, true);
	//ランキングデータファイルが無かった場合は、自動で空ファイル作成
	//ある場合は既存データをKeyValueに落としこむ
	if (FileExists(gs_rankfilepath) == false)
	{
		new Handle:h_createfile = OpenFile(gs_rankfilepath, "w");
		CloseHandle(h_createfile);
	}
	else FileToKeyValues(gh_rankkv, gs_rankfilepath);
}

saveRankData()
{
	KvRewind(gh_rankkv);	//念のため
	KeyValuesToFile(gh_rankkv, gs_rankfilepath);
}

public Action:Cmd_RankReload(client, args)
{
	CloseHandle(gh_rankkv);
	loadRankData();
	return Plugin_Handled;
}

public Action:Cmd_RankFlush(client, args)
{
	saveRankData();
	return Plugin_Handled;
}